"""Test that various classes can be pickled and unpickled."""

import pickle
from thinc.linear.linear import LinearModel


def test_pickle_linear_model():
    model = LinearModel(10)
    model2 = pickle.loads(pickle.dumps(model))



# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
cimport cython
from libc.string cimport memcpy, memset
from libc.math cimport exp, sqrt
from libc.stdlib cimport calloc, malloc, free
import math

from collections import defaultdict
import numpy

from ..typedefs cimport weight_t
from .ops import NumpyOps, CupyOps, add_gradient_noise
from .util import get_array_module
from .._registry import registry


SGD_DEFAULTS = {
    "L2": 1e-4,
    "max_grad_norm": 10,
    "L2_is_weight_decay": False
}


ADAM_DEFAULTS = {
    "learn_rate": 0.001,
    "beta1": 0.9,
    "beta2": 0.999,
    "eps": 1e-08,
    "L2": SGD_DEFAULTS["L2"],
    "max_grad_norm": SGD_DEFAULTS["max_grad_norm"],
    "L2_is_weight_decay": SGD_DEFAULTS["L2_is_weight_decay"],
    "schedules": None
}


@registry.optimizers.register("RAdam.v1")
def create_RAdam(learn_rate=ADAM_DEFAULTS["learn_rate"],
        L2=ADAM_DEFAULTS["L2"],
        beta1=ADAM_DEFAULTS["beta1"],
        beta2=ADAM_DEFAULTS["beta2"],
        eps=ADAM_DEFAULTS["eps"],
        max_grad_norm=ADAM_DEFAULTS["max_grad_norm"],
        L2_is_weight_decay=ADAM_DEFAULTS["L2_is_weight_decay"],
        use_averages=True,
        schedules=None,
        ops=None,
):
    ops = _make_ops(ops)
    return Optimizer(
        ops,
        learn_rate,
        L2=L2,
        beta1=beta1,
        beta2=beta2,
        eps=eps,
        max_grad_norm=max_grad_norm,
        L2_is_weight_decay=L2_is_weight_decay,
        schedules=schedules,
        nesterov=None, lookahead_k=0, lookahead_alpha=0,
        use_averages=True,
        use_radam=True, use_lars=False
    )


@registry.optimizers.register("Adam.v1")
def create_Adam(learn_rate=ADAM_DEFAULTS["learn_rate"],
        L2=ADAM_DEFAULTS["L2"],
        beta1=ADAM_DEFAULTS["beta1"],
        beta2=ADAM_DEFAULTS["beta2"],
        eps=ADAM_DEFAULTS["eps"],
        max_grad_norm=ADAM_DEFAULTS["max_grad_norm"],
        L2_is_weight_decay=ADAM_DEFAULTS["L2_is_weight_decay"],
        use_averages=True,
        ops=None,
        schedules=None
):
    ops = _make_ops(ops)
    return Optimizer(
        ops,
        learn_rate,
        L2=L2,
        beta1=beta1,
        beta2=beta2,
        eps=eps,
        max_grad_norm=max_grad_norm,
        L2_is_weight_decay=L2_is_weight_decay,
        schedules=schedules,
        use_averages=True,
        decay=0.0, decay_steps=0, b1_decay=0, b2_decay=0, 
        nesterov=None, lookahead_k=0, lookahead_alpha=0,
        use_radam=False, use_lars=False
    )


@registry.optimizers.register("SGD.v1")
def create_SGD(learn_rate,
        ops=None,
        L2=SGD_DEFAULTS["L2"],
        max_grad_norm=SGD_DEFAULTS["max_grad_norm"],
        L2_is_weight_decay=SGD_DEFAULTS["L2_is_weight_decay"],
        use_averages=True,
        schedules=None
):
    ops = _make_ops(ops)
    return Optimizer(ops, learn_rate,
        L2=L2, max_grad_norm=max_grad_norm, L2_is_weight_decay=L2_is_weight_decay,
        schedules=schedules, beta1=0.0, beta2=0.0)


class Optimizer(object):
    '''Do various flavours of stochastic gradient descent, with first and
    second order momentum.
    
    Examples
    
    * beta1=0., beta2=0.: "vanilla" SGD
    * beta1=0.9, beta2=0.: "Classic momentum"
    * beta1=0.0, beta2=0.2: RMS prop
    * b1=0.999, b2=0.9: Adam
    '''
    @classmethod
    def from_config(cls, config):
        return registry.make_from_config(config)

    def __init__(self, ops, lr, L2=1e-4, beta1=0.90, beta2=0.999, eps=1e-08, 
                 max_grad_norm=10., gradient_noise=0.0, nesterov=True,
                 L2_is_weight_decay=False, lookahead_k=0, lookahead_alpha=0.5,
                 use_averages=True, use_radam=False, use_lars=False, schedule=None, **_):
        self.ops = ops
        if schedule is None:
            self.schedule = {}
        else:
            self.schedule = dict(schedule)
        self.mom1 = {}
        self.mom2 = {}
        self.slow_weights = {} # For lookahead
        if use_averages:
            self.averages = {}
        else:
            self.averages = None
        self.nr_update = defaultdict(int)
        self.last_seen = defaultdict(int)
        self.max_grad_norm = max_grad_norm
        self.alpha = lr
        self.b1 = beta1
        self.b2 = beta2
        self.gradient_noise = gradient_noise
        self.eps = eps
        self.L2 = L2
        self.nesterov = nesterov
        self.L2_is_weight_decay = L2_is_weight_decay
        self.lookahead_k = lookahead_k
        self.lookahead_alpha = lookahead_alpha
        self.use_radam = use_radam
        # Deprecated
        self.use_lars = use_lars
        self.decay = 0.0
        self.decay_steps = 0
        self.lars_min = 0
        self.lars_max = 10
        self.b1_decay = 0.0
        self.b2_decay = 0.0

    def to_gpu(self):
        self.ops = CupyOps()
        for params in (self.mom1, self.mom2, self.averages):
            for key, value in params.items():
                params[key] = self.ops.xp.asarray(value, dtype=value.dtype)
    
    def to_cpu(self):
        self.ops = NumpyOps()
        for params in (self.mom1, self.mom2, self.averages):
            for key, value in params.items():
                if hasattr(value, 'get'):
                    params[key] = value.get()

    def step_schedules(self):
        for key, schedule in self.schedules.items():
            setattr(self, key, next(schedule))

    @property
    def learn_rate(self):
        return self.alpha

    @learn_rate.setter
    def learn_rate(self, learn_rate):
        self.alpha = learn_rate

    def lr(self, nr_upd):
        alpha = anneal(self.alpha, self.decay, self.decay_steps, nr_upd)
        if self.b1 == 0. or self.b2 == 0.:
            return alpha
        fix1 = 1.- (self.b1 ** nr_upd)
        fix2 = 1.- (self.b2 ** nr_upd)
        return alpha * numpy.sqrt(fix2) / fix1

    def __call__(self, weights, gradient, lr_scale=1., key=None):
        assert len(gradient) >= 1
        xp = get_array_module(weights)
        if xp is not self.ops.xp:
            if xp is numpy:
                self.ops = NumpyOps()
            else:
                self.ops = CupyOps()

        self.nr_update[key] += 1
        nr_upd = self.nr_update[key]
        if self.L2 != 0 and not self.L2_is_weight_decay:
            gradient += self.L2 * weights
        if self.max_grad_norm:
            self.ops.clip_gradient(gradient, self.max_grad_norm)
        if self.gradient_noise:
            add_gradient_noise(gradient, self.gradient_noise, nr_upd)
        if self.use_radam:
            self._radam(xp, weights, gradient, lr_scale, key, nr_upd)
        elif self.b1 > 0. and self.b2 > 0.:
            self._adam(xp, weights, gradient, lr_scale, key, nr_upd)
        elif self.b1 > 0. and not self.nesterov:
            raise NotImplementedError
        elif self.b1 > 0.:
            self._nesterov(xp, weights, gradient, lr_scale, key)
        elif self.b2 > 0.:
            raise NotImplementedError
        else:
            weights -= lr_scale * self.alpha * gradient
        gradient.fill(0.)
        if self.L2 != 0 and self.L2_is_weight_decay:
            weights -= self.L2 * weights
        if self.lookahead_k and self.nr_update[key] % self.lookahead_k == 0:
            if key not in self.slow_weights:
                self.slow_weights[key] = self.ops.allocate((weights.size,), dtype='float32')
            slow = self.slow_weights[key]
            slow += self.lookahead_alpha * (weights - slow)
            weights[:] = slow
        if self.averages is not None:
            if key not in self.averages:
                self.averages[key] = self.ops.allocate((weights.size,), dtype='float32')
            self.ops.update_averages(self.averages[key], weights, nr_upd)

    def _radam(self, xp, weights, gradient, lr_scale, key, nr_upd):
        if key not in self.mom1:
            self.mom1[key] = self.ops.allocate(weights.size)
        if key not in self.mom2:
            self.mom2[key] = self.ops.allocate(weights.size)
 
        beta1 = self.b1
        beta2 = self.b2
        eps = self.eps
        sma_inf = 2 / (1-beta2) - 1

        exp_avg = self.mom1[key]
        exp_avg_sq = self.mom2[key]
        # Decay the first and second moment running average coefficient
        exp_avg *= beta1
        exp_avg += (1-beta1) * gradient
        exp_avg_sq *= beta2
        exp_avg_sq += (1-beta2) * gradient**2
        # Bias correction
        bias_correction1 = 1 - beta1 ** nr_upd
        bias_correction2 = 1 - beta2 ** nr_upd

        # Compute length of SMA
        sma_t = sma_inf - 2 * nr_upd * (1 - bias_correction2) / bias_correction2
        update = self.ops.allocate(weights.shape, dtype="f")
        if sma_t > 4:
            # Variance rectification term
            r_t = math.sqrt((sma_t - 4) * (sma_t - 2) * sma_inf / ((sma_inf - 4) * (sma_inf - 2) * sma_t))
            # Adaptive momentum
            update += r_t * (
                (exp_avg / bias_correction1)
                /
                (self.ops.xp.sqrt(exp_avg_sq / bias_correction2) + eps)
            )
        else:
            # Unadapted momentum
            update += exp_avg / bias_correction1
        if self.use_lars:
            # LARS
            w_norm = self.ops.xp.linalg.norm(weights)
            u_norm = self.ops.xp.linalg.norm(update)
            phi_p = min(max(w_norm, self.lars_min), self.lars_max)
            # Compute the local LR
            if phi_p == 0 or u_norm == 0:
                local_lr = 1
            else:
                local_lr = phi_p / u_norm
            lr = self.alpha * lr_scale * local_lr
        else:
            lr = self.alpha * lr_scale
        weights -= lr * update

    def _nesterov(self, xp, weights, gradient, lr_scale, key):
        # http://cs231n.github.io/neural-networks-3/
        # v_prev = v # back this up
        # v = mu * v - lr * gradient # velocity update stays the same
        # x += -mu * v_prev + (1 + mu) * v # position update changes form
        # Implement this as
        # x += -mu * v
        # v *= mu
        # v -= lr * gradient
        # x += (1+mu) * v
        lr = self.alpha * lr_scale
        if key not in self.mom1:
            self.mom1[key] = self.ops.allocate(weights.size)
        momentum = self.mom1[key]
        weights += -self.b1 * momentum
        momentum *= self.b1
        momentum -= lr * gradient
        weights += (1+self.b1) * momentum

    def _adam(self, xp, weights, gradient, lr_scale, key, nr_upd):
        if key not in self.mom1:
            self.mom1[key] = self.ops.allocate(weights.size)
        if key not in self.mom2:
            self.mom2[key] = self.ops.allocate(weights.size)
        mom1 = self.mom1[key]
        mom2 = self.mom2[key]
        cdef weight_t lr = self.lr(nr_upd)
        cdef weight_t b1 = linear_decay(self.b1, self.b1_decay, nr_upd)
        cdef weight_t b2 = linear_decay(self.b2, self.b2_decay, nr_upd)
        cdef weight_t eps = self.eps
        self.ops.adam(
            weights, gradient, mom1, mom2, b1, b2, eps, lr * lr_scale)
        gradient.fill(0)


def _make_ops(ops):
    if ops == "CupyOps":
        return CupyOps()
    elif ops == "NumpyOps":
        return NumpyOps()
    elif ops is None:
        from ._classes.model import Model
        return Model.ops
    else:
        return ops


# These are deprecated

def Adam(*args, **kwargs):
    return Optimizer(*args, **kwargs)

def SGD(*args, **kwargs):
    kwargs.setdefault('beta1', 0.)
    kwargs.setdefault('beta2', 0.)
    return Optimizer(*args, **kwargs)


def linear_decay(rate, decay, nr_upd):
    return rate * 1./(1. + decay * nr_upd)


def anneal(rate, decay, decay_steps, nr_upd):
    if decay == 0.0:
        return rate
    else:
        return rate * decay ** (nr_upd / decay_steps)



# inspired from:

# https://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/
# https://github.com/pypa/warehouse/blob/master/warehouse/__about__.py

__name__ = "thinc"
__version__ = "7.4.4"
__summary__ = "Practical Machine Learning for NLP"
__uri__ = "https://github.com/explosion/thinc"
__author__ = "Matthew Honnibal"
__email__ = "matt@explosion.ai"
__license__ = "MIT"
__title__ = "thinc"
__release__ = True



# coding: utf8
from __future__ import unicode_literals

from .neural._classes.convolution import ExtractWindow  # noqa: F401
from .neural._classes.attention import ParametricAttention  # noqa: F401
from .neural._classes.rnn import LSTM, BiLSTM  # noqa: F401
from .neural._classes.multiheaded_attention import MultiHeadedAttention
from .neural._classes.multiheaded_attention import prepare_self_attention
from ._registry import registry


@registry.layers.register("ExtractWindow.v1")
def make_ExtractWindow(window_size):
    return ExtractWindow(nW=window_size)


@registry.layers.register("ParametricAttention.v1")
def make_ParametricAttention(outputs):
    return ParametricAttention(nO=outputs)



# coding: utf8
from __future__ import unicode_literals

from thinc import about


def test_about_attrs():
    """Test that about.py imports correctly and has expected attributes."""
    about.__name__
    about.__version__
    about.__summary__
    about.__uri__
    about.__author__
    about.__email__
    about.__license__
    about.__title__
    about.__release__



# cython: infer_types=True
# cython: cdivision=True
# cython: bounds_check=False
# cython: wraparound=False
from murmurhash.mrmr cimport hash32
cimport numpy as np
from libc.stdint cimport uint64_t, int32_t, uint32_t
from ..neural._classes.model import Model
from .. import describe
from ..describe import Dimension, Synapses, Biases, Gradient
from ..neural.util import is_cupy_array, is_numpy_array

# Sigh, these stuff up pickling if they're lambdas...

def _get_W_shape(obj):
    return (obj.nO * obj.length,)

def _init_W(W, ops):
    W.fill(0.)

def _get_bias_shape(obj):
    return (obj.nO,)


@describe.attributes(
    nO=Dimension("Output size"),
    length=Dimension("Weights length"),
    W=Synapses("Weights matrix", _get_W_shape, _init_W),
    b=Biases("Biases", _get_bias_shape),
    d_W=Gradient("W"),
    d_b=Gradient("b"),
)
class LinearModel(Model):
    name = 'linear'
    def __init__(self, nO, length=2**18, **kwargs):
        Model.__init__(self, **kwargs)
        self.nO = nO
        self.length = length

    def begin_update(self, keys_values_lengths, drop=0.):
        keys, values, lengths = keys_values_lengths
        if is_cupy_array(keys):
            return self._begin_gpu_update(keys, values, lengths, drop=drop)
        else:
            return self._begin_cpu_update(keys, values, lengths, drop=drop)

    def _begin_gpu_update(self, keys, values, lengths, drop):
        # Currently we don't have a GPU-compatible implementation of this function :(
        # It sucks, but at least we can get the correct result by copying to CPU.
        cpu_keys = keys.get()
        cpu_values = values.get()
        cpu_lengths = lengths.get()
        return self._begin_cpu_update(cpu_keys, cpu_values, cpu_lengths, drop=drop)

    def _begin_cpu_update(self, uint64_t[::1] keys, np.ndarray values_, long[::1] lengths, drop):
        if drop is not None:
            drop *= self.drop_factor
        mask = self.ops.get_dropout_mask((values_.shape[0],), drop)
        cdef float[::1] values
        if mask is not None:
            values = values_ * mask
        else:
            values = values_
        cdef float[:, ::1] scores = self.ops.allocate((len(lengths), self.nO)) + self.b
        cdef float[::1] weights = self.W
        set_scoresC(&scores[0, 0],
            &keys[0], &values[0], &lengths[0],
            lengths.shape[0], self.nO,
            &weights[0], self.length)
        return scores, _finish_linear_update(self, keys, values, lengths)


class _finish_linear_update(object):
    """Move this out of a closure, into its own callable object, to avoid
    pickling errors :(."""
    def __init__(self, layer, keys, values, lengths):
        self.layer = layer
        self.keys = keys
        self.values = values
        self.lengths = lengths

    def __call__(self, float[:, ::1] d_scores, sgd=None):
        cdef float[::1] d_weights = self.layer.d_W
        cdef float[::1] d_bias = self.layer.d_b
        cdef uint64_t[::1] keys = self.keys
        cdef float[::1] values = self.values
        cdef long[::1] lengths = self.lengths
        set_gradientC(&d_weights[0],
            &keys[0], &values[0], &lengths[0],
            lengths.shape[0], self.layer.nO,
            &d_scores[0,0], self.layer.length)
        cdef int i, j
        for i in range(d_scores.shape[0]):
            for j in range(d_scores.shape[1]):
                d_bias[j] += d_scores[i, j]
        if sgd is not None:
            sgd(self.layer._mem.weights, self.layer._mem.gradient, key=self.layer.id)
        return None


cdef void set_scoresC(float* scores,
        const uint64_t* keys, const float* values, const long* lengths,
        int batch_size, int nr_out,
        const float* weights, int nr_weight) nogil:
    cdef uint32_t idx1, idx2
    cdef uint32_t hash1, hash2
    for length in lengths[:batch_size]:
        for i in range(length):
            hash1 = hash32(<void*>&keys[i], sizeof(keys[i]), 0)
            hash2 = hash32(<void*>&keys[i], sizeof(keys[i]), 1)
            idx1 = hash1 & (nr_weight-1)
            idx2 = hash2 & (nr_weight-1)
            value = values[i]
            for clas in range(nr_out):
                scores[clas] += weights[idx1 + clas] * value
                scores[clas] += weights[idx2 + clas] * value
        scores += nr_out
        keys += length
        values += length


cdef void set_gradientC(float* d_weights,
        const uint64_t* keys, const float* values, const long* lengths,
        int batch_size, int nr_out,
        const float* d_scores, int nr_weight) nogil:
    cdef uint32_t idx1, idx2
    cdef uint32_t hash1, hash2
    for length in lengths[:batch_size]:
        for i in range(length):
            hash1 = hash32(<void*>&keys[i], sizeof(keys[i]), 0)
            hash2 = hash32(<void*>&keys[i], sizeof(keys[i]), 1)
            idx1 = hash1 & (nr_weight-1)
            idx2 = hash2 & (nr_weight-1)
            value = values[i]
            for clas in range(nr_out):
                d_weights[idx1 + clas] += d_scores[clas] * value
                d_weights[idx2 + clas] += d_scores[clas] * value
        d_scores += nr_out
        keys += length
        #values += length



# coding: utf8
"""Generators that provide different rates, schedules, decays or series."""
from __future__ import unicode_literals, division
import numpy
from ._registry import registry


@registry.schedules.register("constant_then.v1")
def constant_then(rate, steps, schedule):
    """Yield a constant rate for N steps, before starting a schedule."""
    for i in range(steps):
        yield rate
    for value in schedule:
        yield value


@registry.schedules.register("constant.v1")
def constant(rate):
    while True:
        yield rate


@registry.schedules.register("decaying.v1")
def decaying(base_rate, decay, t=0):
    """Yield an infinite series of linearly decaying values,
    following the schedule:

        base_rate * 1/(1+decay*t)

    Example:

        >>> learn_rates = linear_decay(0.001, 1e-4)
        >>> next(learn_rates)
        0.001
        >>> next(learn_rates)
        0.00999
    """
    while True:
        yield base_rate * (1.0 / (1.0 + decay * t))
        t += 1


@registry.schedules.register("compounding.v1")
def compounding(start, stop, compound, t=0.0):
    """Yield an infinite series of compounding values. Each time the
    generator is called, a value is produced by multiplying the previous
    value by the compound rate.

    EXAMPLE:
      >>> sizes = compounding(1., 10., 1.5)
      >>> assert next(sizes) == 1.
      >>> assert next(sizes) == 1 * 1.5
      >>> assert next(sizes) == 1.5 * 1.5
    """
    curr = float(start)
    while True:
        yield _clip(curr, start, stop)
        curr *= compound


def _clip(value, start, stop):
    return max(value, stop) if (start > stop) else min(value, stop)



@registry.schedules.register("slanted_triangular.v1")
def slanted_triangular(max_rate, num_steps, cut_frac=0.1, ratio=32, decay=1, t=0.0):
    """Yield an infinite series of values according to Howard and Ruder's
    "slanted triangular learning rate" schedule.
    """
    cut = int(num_steps * cut_frac)
    while True:
        t += 1
        if t < cut:
            p = t / cut
        else:
            p = 1 - ((t - cut) / (cut * (1 / cut_frac - 1)))
        learn_rate = max_rate * (1 + p * (ratio - 1)) * (1 / ratio)
        yield learn_rate


@registry.schedules.register("warmup_linear.v1")
def warmup_linear(initial_rate, warmup_steps, total_steps):
    """Generate a series, starting from an initial rate, and then with a warmup
    period, and then a linear decline. Used for learning rates.
    """
    step = 0
    while True:
        if step < warmup_steps:
            factor = step / max(1, warmup_steps)
        else:
            factor = max(
                0.0, (total_steps - step) / max(1.0, total_steps - warmup_steps)
            )
        yield factor * initial_rate
        step += 1


@registry.schedules.register("cyclic_triangular.v1")
def cyclic_triangular(min_lr, max_lr, period):
    it = 1
    while True:
        # https://towardsdatascience.com/adaptive-and-cyclical-learning-rates-using-pytorch-2bf904d18dee
        cycle = numpy.floor(1 + it / (2 * period))
        x = numpy.abs(it / period - 2 * cycle + 1)
        relative = max(0, 1 - x)
        yield min_lr + (max_lr - min_lr) * relative
        it += 1


# Deprecated

def annealing(rate, decay, decay_steps, t=0.0):
    while True:
        if decay == 0.0:
            yield rate
        else:
            yield rate * decay ** (t / decay_steps)
            t += 1


def annealing_cos(start, end, step=0.001):
    pct = step
    while True:
        cos_out = numpy.cos(numpy.pi * pct) + 1
        yield end + (start-end)/2 * cos_out
        pct += step





# coding: utf8
from __future__ import unicode_literals

import pytest
import numpy
from numpy.testing import assert_allclose

from thinc.neural._classes.feed_forward import FeedForward
from thinc.neural._classes.affine import Affine
from thinc.neural._classes.relu import ReLu


@pytest.fixture
def model1(nH, nI):
    model = ReLu(nH, nI)
    return model


@pytest.fixture
def model2(nO, nH):
    model = Affine(nO, nH)
    return model


@pytest.fixture
def input_data(nB, nI):
    return numpy.ones((nB, nI), dtype="f") + 1.0


@pytest.fixture
def gradient_data(nB, nO):
    return numpy.zeros((nB, nO), dtype="f") - 1.0


@pytest.fixture
def model(model1, model2):
    return FeedForward((model1, model2))


def get_expected_predict(input_data, Ws, bs):
    X = input_data
    for i, (W, b) in enumerate(zip(Ws, bs)):
        X = numpy.ascontiguousarray(X)
        if i > 0:
            X *= X > 0
        X = numpy.tensordot(X, W, axes=[[1], [1]]) + b
    return X


def numeric_gradient(predict, weights, epsilon=1e-4):
    out1 = predict(weights + epsilon)
    out2 = predict(weights - epsilon)
    return (out1 - out2) / (2 * epsilon)


def test_models_have_shape(model1, model2, nI, nH, nO):
    assert model1.W.shape == (nH, nI)
    assert model1.b.shape == (nH,)
    assert model2.W.shape == (nO, nH)
    assert model2.b.shape == (nO,)


def test_model_shape(model, model1, model2, nI, nH, nO):
    assert model.input_shape == model1.input_shape
    assert model.output_shape == model2.output_shape


def test_predict_and_begin_update_match(model, model1, model2, input_data):
    model = FeedForward((model1, model2))
    via_predict = model.predict(input_data)
    via_update, _ = model.begin_update(input_data)
    assert_allclose(via_predict, via_update)
    expected = get_expected_predict(
        input_data, [model1.W, model2.W], [model1.b, model2.b]
    )
    assert_allclose(via_update, expected, atol=1e-2, rtol=1e-4)


class GradientSpy(object):
    def __init__(self):
        self.weights = None
        self.d_weights = None

    def __call__(self, weights, grad):
        self.weights = weights
        self.d_weights = grad


def test_gradient(model, input_data, nB, nH, nI, nO):
    truth = numpy.zeros((nB, nO), dtype="float32")
    truth[0] = 1.0

    guess, backprop = model.begin_update(input_data)
    backprop(guess - truth)

    for layer in model._layers:

        def predict(i, update):
            layer._mem.weights[i] += update
            X = model.predict(input_data)
            layer._mem.weights[i] -= update
            return X

        agrad = layer._mem.gradient.copy()
        ngrad = get_numeric_gradient(predict, layer._mem.weights.size, truth)
        assert_allclose(agrad, ngrad, atol=0.2, rtol=0.2)


def get_numeric_gradient(predict, n, target):
    gradient = numpy.zeros(n)
    for i in range(n):
        out1 = predict(i, 1e-4)
        out2 = predict(i, -1e-4)

        err1 = _get_loss(out1, target)
        err2 = _get_loss(out2, target)
        gradient[i] = (err1 - err2) / (2 * 1e-4)
    return gradient


def _get_loss(truth, guess):
    return numpy.sum(numpy.sum(0.5 * numpy.square(truth - guess), 1))



from __future__ import unicode_literals, print_function

from .model import Model
from .affine import Affine
from ...api import with_reshape, layerize, wrap
from ..util import get_array_module
import numpy as np
import copy
import math


def prepare_self_attention(affine, window=None, nM=300, nH=6):
    nD = nM // nH
    if window is not None:
        get_mask = window_mask(window)
    else:
        get_mask = None
    def qkv_sa_forward(Xs, drop=0.0):
        X = affine.ops.flatten(Xs)
        lengths = [len(x) for x in Xs]
        QKV, get_dX = affine.begin_update(X, drop=drop)
        Qs, Ks, Vs = _split_seqs(QKV, lengths, nH, nD)

        def qkv_sa_backward(dQs_dKs_dVs, sgd=None):
            dQs, dKs, dVs = dQs_dKs_dVs
            dQKV = _join_seqs(dQs, dKs, dVs, nH, nD)
            dX = get_dX(dQKV, sgd=sgd)
            return affine.ops.unflatten(dX, lengths)
        if get_mask is not None:
            xp = get_array_module(X)
            masks = [get_mask(xp, length, length) for length in lengths]
        else:
            masks = [None for _ in lengths]
        return (Qs, Ks, Vs, masks), qkv_sa_backward
    return wrap(qkv_sa_forward, affine)


def window_mask(n):
    def get_mask(xp, nX, nY):
        mask = xp.zeros((nX, nY), dtype='f')
        for i in range(nX):
            mask[i, i-n:i+n] = 1
        return mask
    return get_mask


def _split_seqs(QKV, lengths, nH, nD):
    assert sum(lengths) == QKV.shape[0], (sum(lengths), QKV.shape[0])
    Qs = []
    Ks = []
    Vs = []
    i = 0
    xp = get_array_module(QKV)
    for length in lengths:
        qkv = QKV[i:i+length]
        qkv = qkv.reshape((length, 3, nH*nD))
        queries = xp.ascontiguousarray(qkv[:, 0])
        keys = xp.ascontiguousarray(qkv[:, 1])
        values = xp.ascontiguousarray(qkv[:, 2])
        Qs.append(queries.reshape((-1, nH, nD)))
        Ks.append(keys.reshape((-1, nH, nD)))
        Vs.append(values.reshape((-1, nH, nD)))
        i += length
    return Qs, Ks, Vs


def _join_seqs(Qs, Ks, Vs, nH, nD):
    xp = get_array_module(Qs[0])
    Q = xp.vstack(Qs).reshape((-1, nH*nD))
    K = xp.vstack(Ks).reshape((-1, nH*nD))
    V = xp.vstack(Vs).reshape((-1, nH*nD))
    assert Q.shape[0] == K.shape[0] == V.shape[0]
    return xp.hstack((Q, K, V))


class MultiHeadedAttention(Model):
    """Multi-headed attention. Requires a preprocessor to prepare (Qs, Ks, Vs, masks)
    triples, such as the prepare_self_attention() preprocessor. A layer
    should run after this to do the projection as well."""
    def __init__(self):
        Model.__init__(self)

    def begin_update(self, Qs_Ks_Vs_masks, drop=0.0):
        Qs, Ks, Vs, masks = Qs_Ks_Vs_masks
        if masks is None:
            masks = [None for _ in Qs]
        assert len(Qs) == len(Ks) == len(Vs)
        return self._attend_seqs(Qs, Ks, Vs, masks)
        
    def _attend_seqs(self, Qs, Ks, Vs, masks):
        outputs = []
        backprops = []
        assert len(Qs) == len(Ks) == len(Vs) == len(masks), (len(Qs), len(Ks), len(Vs), len(masks))
        for Q, K, V, mask in zip(Qs, Ks, Vs, masks):
            output, backprop = self._attend(Q, K, V, mask)
            outputs.append(output)
            backprops.append(backprop)
            assert output.shape[0] == Q.shape[0]

        def backprop_attend_seqs(d_outputs, sgd=None):
            dQs = []
            dKs = []
            dVs = []
            for d_output, backprop in zip(d_outputs, backprops):
                dQ, dK, dV = backprop(d_output, sgd=sgd)
                dQs.append(dQ)
                dKs.append(dK)
                dVs.append(dV)
            return dQs, dKs, dVs
        assert len(outputs) == len(Qs), len(Qs)
        return outputs, backprop_attend_seqs

    def _attend(self, Q, K, V, mask):
        """
        Compute attention on a (query, key, value) triplet.
        The similarity of the (Q, K) pairs are used to
        compute an attention matrix, which is used to rescale
        V.
        """
        attn, get_dQ_dK = self._get_attn_weights(Q, K, mask)
        output, get_d_attn_dV = self._apply_attn(attn, V)

        def backprop_attend(d_output, sgd=None):
            d_attn, dV = get_d_attn_dV(d_output)
            dQ, dK = get_dQ_dK(d_attn)
            return (dQ, dK, dV)

        return output, backprop_attend

    def _get_attn_weights(self, Q0, K0, mask):
        nQ, nK, nH, nD = (Q0.shape[0], K0.shape[0], Q0.shape[1], Q0.shape[2])
        assert Q0.shape == (nQ, nH, nD)
        assert K0.shape == (nK, nH, nD)
        sqrtM = self.ops.xp.sqrt(nH*nD).astype("f")
        Q1 = _trans(Q0, 1, 0, 2)
        assert Q1.shape == (nH, nQ, nD)
        K1 = _trans(K0, 1, 2, 0)
        assert K1.shape == (nH, nD, nK)
        K1 /= sqrtM
        attn0 = self.ops.matmul(Q1, K1)
        assert attn0.shape == (nH, nQ, nK)
        attn1, backprop_mask = self._apply_mask(attn0, mask)
        attn2 = self.ops.softmax(attn1, axis=-1)
        assert attn2.shape == (nH, nQ, nK)

        def backprop_attn1(d_attn2, sgd=None):
            assert d_attn2.shape == (nH, nQ, nK)
            d_attn1 = self.ops.backprop_softmax(attn2, d_attn2, axis=-1)
            d_attn0 = backprop_mask(d_attn1)
            assert d_attn0.shape == (nH, nQ, nK)
            dQ1 = self.ops.matmul(d_attn0, _trans(K1, 0, 2, 1))
            assert dQ1.shape == (nH, nQ, nD)
            dK1 = self.ops.matmul(_trans(Q1, 0, 2, 1), d_attn0)
            assert dK1.shape == (nH, nD, nK)
            dK0 = _trans(dK1, 2, 0, 1)
            dK0 /= sqrtM
            assert dK0.shape == (nK, nH, nD)
            dQ0 = _trans(dQ1, 1, 0, 2)
            assert dQ0.shape == (nQ, nH, nD)
            return dQ0, dK0

        return attn2, backprop_attn1

    def _apply_mask(self, attn, mask):
        def backprop_apply_mask(d_attn, sgd=None):
            if mask is None:
                return d_attn
            else:
                return d_attn * mask

        if mask is None:
            return attn, backprop_apply_mask
        else:
            return attn - (1-mask)*1e9, backprop_apply_mask

    def _apply_attn(self, attn, V0):
        """ Multiplication with values """
        nH, nQ, nV = attn.shape
        nD = V0.shape[-1]
        assert V0.shape == (nV, nH, nD)
        V1 = _trans(V0, 1, 0, 2)
        assert V1.shape == (nH, nV, nD)
        # (nH, nQ, nV) @ (nH, nV, nD) = (nH, nQ, nD)
        S0 = self.ops.matmul(attn, V1)
        assert S0.shape == (nH, nQ, nD)
        S1 = _trans(S0, 1, 0, 2)
        assert S1.shape == (nQ, nH, nD)
        S2 = S1.reshape((nQ, nH*nD))

        def backprop_apply_attn(dS2):
            assert dS2.shape == (nQ, nH*nD)
            dS1 = dS2.reshape((nQ, nH, nD))
            dS0 = dS1.transpose((1, 0, 2))
            assert dS0.shape == (nH, nQ, nD)
            dS0 = self.ops.xp.ascontiguousarray(dS0)
            # (nH, nQ, nD) @ (nH, nD, nV) --> (nH, nQ, nV)
            d_attn = self.ops.matmul(dS0, _trans(V1, 0, 2, 1))
            assert d_attn.shape == (nH, nQ, nV)
            # (nH, nV, nQ) @ (nH, nQ, nD) --> (nH, nV, nD)
            dV1 = self.ops.matmul(_trans(attn, 0, 2, 1), dS0)
            assert dV1.shape == (nH, nV, nD)
            dV0 = dV1.transpose((1, 0, 2))
            assert dV0.shape == (nV, nH, nD)
            return d_attn, dV0

        return S2, backprop_apply_attn

def _trans(X, *order):
    """Transpose and make contiguous"""
    xp = get_array_module(X)
    return xp.ascontiguousarray(X.transpose(order))



# coding: utf8
from __future__ import unicode_literals

# Necessary for some side-effects in Cython. Not sure I understand.
import numpy  # noqa: F401

from .about import __name__, __version__  # noqa: F401
from ._registry import registry



# coding: utf8
from __future__ import unicode_literals
import numpy

from ... import describe
from .model import Model
from ...describe import Dimension, Synapses, Biases, Gradient


def _set_dimensions_if_needed(model, X, y=None):
    if model.nI is None:
        model.nI = X.shape[1]
    if model.nO is None and y is not None:
        if len(y.shape) == 2:
            model.nO = y.shape[1]
        else:
            model.nO = int(y.max()) + 1


@describe.on_data(_set_dimensions_if_needed)
@describe.attributes(
    nB=Dimension("Batch size"),
    nI=Dimension("Input size"),
    nO=Dimension("Output size"),
    W=Synapses(
        "Weights matrix",
        lambda obj: (obj.nO, obj.nI),
        lambda W, ops: ops.xavier_uniform_init(W),
    ),
    b=Biases("Bias vector", lambda obj: (obj.nO,)),
    d_W=Gradient("W"),
    d_b=Gradient("b"),
)
class Mish(Model):
    """Dense layer with mish activation.
    
    https://arxiv.org/pdf/1908.08681.pdf
    """
    name = "mish"

    @property
    def input_shape(self):
        return (self.nB, self.nI)

    @property
    def output_shape(self):
        return (self.nB, self.nO)

    def __init__(self, nO=None, nI=None, **kwargs):
        Model.__init__(self, **kwargs)
        self.nO = nO
        self.nI = nI
        self.drop_factor = kwargs.get("drop_factor", 1.0)

    def predict(self, X):
        Y = self.ops.affine(self.W, self.b, X)
        Y = self.ops.mish(Y)
        return Y

    def begin_update(self, X, drop=0.0):
        if drop is None:
            return self.predict(X), None
        Y1 = self.ops.affine(self.W, self.b, X)
        Y2 = self.ops.mish(Y1)
        drop *= self.drop_factor
        Y3, bp_dropout = self.ops.dropout(Y2, drop)

        def finish_update(dY2, sgd=None):
            dY1 = self.ops.backprop_mish(dY2, Y1)
            self.ops.gemm(dY1, X, trans1=True, out=self.d_W)
            self.d_b += dY1.sum(axis=0)
            dX = self.ops.gemm(dY1, self.W)
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=self.id)
            return dX

        return Y3, bp_dropout(finish_update)
